from contextlib import contextmanager

import pytest
from opperai import Client
from opperai.types import (
    CacheConfiguration,
    ChatPayload,
    Function,
    Message,
    MessageContent,
)
from opperai.types.exceptions import StructuredGenerationError


@contextmanager
def _function(desc: Function, c: Client):
    function = c.functions.create(desc)
    yield function
    c.functions.delete(id=function.id)


def test_create_function(vcr_cassette, client: Client):
    with _function(
        Function(
            path="test/sdk/test_create_function",
            description="Test function",
            instructions="Do something",
        ),
        client,
    ) as function:
        assert function is not None


def test_get_by_id(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_get_by_id",
            description="Test function",
            instructions="Do something",
        ),
        client,
    ) as function:
        f_by_id = client.functions.get(id=function.id)
        assert f_by_id.path == "test/sdk/test_get_by_id"


def test_get_by_path(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_get_by_path",
            description="Test function",
            instructions="Do something",
        ),
        client,
    ) as function:
        f_by_path = client.functions.get(path=function.path)
        assert f_by_path.path == "test/sdk/test_get_by_path"


def test_update_function(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_update_function",
            description="Test function",
            instructions="Do something",
        ),
        client,
    ) as function:
        f = client.functions.get(id=function.id)
        f.instructions = "Do something else"
        f1 = client.functions.update(f)
        f2 = client.functions.get(id=f1.id)
        assert f1 == f2
        assert f2.instructions == "Do something else"


def test_delete_function_by_id(client: Client, vcr_cassette):
    function = client.functions.create(
        Function(
            path="test/sdk/test_delete_function_by_id",
            description="Test function",
            instructions="Do something",
        )
    )
    assert function is not None
    client.functions.delete(id=function.id)
    f = client.functions.get(id=function.id)
    assert f is None


def test_delete_function_by_path(client: Client, vcr_cassette):
    function = client.functions.create(
        Function(
            path="test/sdk/test_delete_function_by_path",
            description="Test function",
            instructions="Do something",
        )
    )
    assert function is not None
    client.functions.delete(path=function.path)
    f = client.functions.get(path=function.path)
    assert f is None


def test_image_url(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_image",
            instructions="describe the image",
            model="openai/gpt4-turbo",
        ),
        client,
    ) as function:
        f = client.functions.get(id=function.id)
        resp = client.functions.chat(
            f.path,
            ChatPayload(
                messages=[
                    Message(
                        role="user",
                        content=[
                            MessageContent.image_url(
                                url="https://tf-cmsv2-smithsonianmag-media.s3.amazonaws.com/filer/e1/d3/e1d3fee9-9aaa-4599-ba67-1c71a6d0ed03/1200px-seymouria_fossil.jpg"
                            )
                        ],
                    )
                ]
            ),
        )

        assert "fossil" in resp.message.lower()


def test_image_file(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_image",
            instructions="describe the image",
            model="openai/gpt4-turbo",
        ),
        client,
    ) as function:
        f = client.functions.get(id=function.id)
        resp = client.functions.chat(
            f.path,
            ChatPayload(
                messages=[
                    Message(
                        role="user",
                        content=[
                            MessageContent.image(
                                path="tests/fixtures/images/fossil.jpg"
                            )
                        ],
                    )
                ]
            ),
        )

        assert "fossil" in resp.message.lower()


def test_chat(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_chat",
            description="Translate to French",
            instructions="Translate to French",
        ),
        client,
    ) as function:
        f = client.functions.get(id=function.id)
        resp = client.functions.chat(
            f.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )

        assert "bonjour" in resp.message.lower()


def test_chat_stream(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_sync_chat_stream",
            description="Translate to French",
            instructions="Translate to French",
        ),
        client,
    ) as function:
        f = client.functions.get(id=function.id)
        gen = client.functions.chat(
            f.path,
            ChatPayload(messages=[Message(role="user", content="hello")]),
            stream=True,
        )

        resp = "".join([message.delta for message in gen if message.delta])

        assert "bonjour" in resp.lower()


def test_create_function_with_cache(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_create_function_with_cache",
            description="Test function",
            instructions="Do something",
            cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
        ),
        client,
    ) as function:
        res = client.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached

        res = client.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert res.cached


def test_create_function_with_cache_flush(client: Client, vcr_cassette):
    with _function(
        Function(
            path="test/sdk/test_create_function_with_cache_sync_flush",
            description="Test function",
            instructions="Do something",
            cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
        ),
        client,
    ) as function:
        res = client.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached

        res = client.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert res.cached

        client.functions.flush_cache(id=function.id)
        res = client.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached


def test_failed_structured_generation(client: Client, vcr_cassette):
    with _function(
        Function(
            model="mistral/mistral-tiny-eu",
            path="test/sdk/test_failed_structured_generation",
            description="test structured generation exception",
            instructions="You translate the incoming text to french",
            out_schema={
                "type": "object",
                "properties": {
                    "universityName": {"type": "string"},
                    "location": {"type": "string"},
                    "departments": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "departmentName": {"type": "string"},
                                "head": {"type": "string"},
                                "courses": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "courseId": {"type": "string"},
                                            "courseName": {"type": "string"},
                                            "credits": {
                                                "type": "integer",
                                                "minimum": 1,
                                                "maximum": 10,
                                            },
                                        },
                                        "required": [
                                            "courseId",
                                            "courseName",
                                            "credits",
                                        ],
                                    },
                                },
                            },
                            "required": ["departmentName", "head", "courses"],
                        },
                    },
                },
                "required": ["universityName", "location", "departments"],
            },
        ),
        client,
    ) as function:
        with pytest.raises(StructuredGenerationError):
            client.functions.chat(
                function.path,
                ChatPayload(messages=[Message(role="user", content="hello")]),
            )



import itertools
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Union

from opperai._client import Client
from opperai.functions.decorator._schemas import type_to_json_schema
from opperai.types import ChatPayload, Message, StreamingChunk
from opperai.types import Function as FunctionModel
from opperai.types import FunctionResponse as FunctionResponseModel
from pydantic import PrivateAttr

from ..spans.spans import Span


class FunctionResponse(FunctionResponseModel):
    _client: Client = PrivateAttr()

    def __init__(self, client: Client = None, **kwargs):
        super().__init__(**kwargs)
        if not client:
            client = Client()
        self._client = client

    @property
    def span(self):
        return Span(self._client, self.span_id)


class StreamingResponse:
    _client: Client = None
    _head: StreamingChunk = None
    _stream: Iterator[StreamingChunk] = None
    context: List[Dict[str, Any]] = []
    span: Span = None

    def __init__(
        self,
        client: Client = None,
        stream: Iterator[StreamingChunk] = None,
    ):
        if not client:
            client = Client()
        self._client = client

        _stream, _head = itertools.tee(stream)
        self._stream = _stream

        head = next(_head)
        self.span = Span(self._client, head.span_id)
        self.context = head.context if head.context else []

    @property
    def deltas(self) -> Iterator[str]:
        for chunk in self._stream:
            if chunk.delta is not None:
                yield chunk.delta


@dataclass
class Function:
    _client: Client
    _function: FunctionModel

    def chat(
        self,
        messages: List[Message],
        parent_span_id: Optional[str] = None,
        stream: Optional[bool] = False,
    ) -> Union[FunctionResponse, StreamingResponse]:
        if stream:
            if self._function.out_schema is not None:
                raise ValueError(
                    "Cannot stream output for functions structured response"
                )

            response: Iterator[StreamingChunk] = self._client.functions.chat(
                function_path=self._function.path,
                data=ChatPayload(messages=messages, parent_span_uuid=parent_span_id),
                stream=stream,
            )

            return StreamingResponse(client=self._client, stream=response)

        response: FunctionResponseModel = self._client.functions.chat(
            function_path=self._function.path,
            data=ChatPayload(messages=messages, parent_span_uuid=parent_span_id),
            stream=stream,
        )

        return FunctionResponse(client=self._client, **response.model_dump())

    def delete(self) -> bool:
        return self._client.functions.delete(id=self._function.id)

    def flush_cache(self) -> bool:
        return self._client.functions.flush_cache(id=self._function.id)

    def update(self, **kwargs) -> "Function":
        updated = self._function.model_dump(exclude_none=True)
        if "input_type" in kwargs and kwargs["input_type"] is not None:
            kwargs["input_schema"] = type_to_json_schema(kwargs["input_type"])
            del kwargs["input_type"]
        if "output_type" in kwargs and kwargs["output_type"] is not None:
            kwargs["out_schema"] = type_to_json_schema(kwargs["output_type"])
            del kwargs["output_type"]

        for key, value in kwargs.items():
            updated[key] = value

        updated_model = FunctionModel.model_validate(updated)
        updated_function = self._client.functions.update(
            id=self._function.id, function=updated_model
        )
        self._function = updated_function

        return self


@dataclass
class Functions:
    _client: Client = None

    def __init__(self, client: Client = None):
        if client is None:
            client = Client()

        self._client = client

    def create(
        self,
        path: str,
        instructions: str,
        description: Optional[str] = None,
        input_type: Optional[Any] = None,
        output_type: Optional[Any] = None,
        model: Optional[str] = None,
    ) -> Function:
        try:
            function = self.get(path=path)
            if function:
                return function.update(
                    instructions=instructions,
                    description=description,
                    input_type=input_type,
                    output_type=output_type,
                    model=model,
                )
        except Exception:
            pass

        if input_type:
            input_schema = type_to_json_schema(input_type)
        if output_type:
            output_schema = type_to_json_schema(output_type)

        function = self._client.functions.create(
            FunctionModel(
                path=path,
                instructions=instructions,
                description=description,
                input_schema=input_schema if input_type else None,
                out_schema=output_schema if output_type else None,
                model=model,
            )
        )

        return Function(self._client, function)

    def get(self, id: int = None, path: str = None) -> Optional[Function]:
        if id is not None:
            function = self._client.functions.get(id=id)
        elif path is not None:
            function = self._client.functions.get(path=path)
        else:
            raise ValueError("Either id or name must be provided")

        if not function:
            return None

        return Function(self._client, function)

    def delete(self, id: int = None, path: str = None) -> bool:
        if id is not None:
            return self._client.functions.delete(id=id)
        elif path is not None:
            return self._client.functions.delete(path=path)
        else:
            raise ValueError("Either id or name must be provided")



import asyncio
import contextlib
import inspect
import json
import os
import threading
from functools import wraps
from typing import Any, List, Tuple, get_args, get_origin, get_type_hints

from opperai._client import AsyncClient, Client
from opperai.core.spans import get_current_span_id
from opperai.core.utils import convert_function_call_to_json
from opperai.functions.async_functions import AsyncFunctionResponse
from opperai.functions.functions import FunctionResponse
from opperai.types import CacheConfiguration, ChatPayload, Function, Message
from pydantic import BaseModel

from ._schemas import get_output_schema

_thread_local = threading.local()


class FunctionWrapper:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs) -> Any:
        answer, _ = self.func(*args, **kwargs)
        return answer

    def call(self, *args, **kwargs) -> Tuple[Any, FunctionResponse]:
        return self.func(*args, **kwargs)


class AsyncFunctionWrapper:
    def __init__(self, func):
        self.func = func

    async def __call__(self, *args, **kwargs) -> Any:
        answer, _ = await self.func(*args, **kwargs)
        return answer

    async def call(self, *args, **kwargs) -> Tuple[Any, AsyncFunctionResponse]:
        return await self.func(*args, **kwargs)


@contextlib.contextmanager
def span_id_context():
    _thread_local.span_id = None
    try:
        yield
    finally:
        del _thread_local.span_id


def get_last_span_id() -> str:
    """Retrieve the last span ID from thread-local storage."""
    return getattr(_thread_local, "span_id", None)


def fn(
    _func=None,
    *,
    path=None,
    client=None,
    json_encoder=None,
    model=None,
    few_shot=None,
    few_shot_count=None,
    cache_config: CacheConfiguration = None,
):
    """Decorator to to create a function in OpperAI's API

    This decorator can automatically handle both synchronous and asynchronous functions,
    wrapping them to perform necessary setup before invoking the OpperAI API. It allows
    for the specification of various parameters to customize the API request, including
    the use of few-shot learning, caching, and custom JSON encoding.

    Parameters:
    - path (str, optional): The API path for the function. Defaults to the decorated function's name.
    - client (Client or AsyncClient, optional): The OpperAI client instance to use. If not provided, a new instance is created.
    - json_encoder (JSONEncoder, optional): Custom JSON encoder for serializing the request payload.
    - model (str, optional): The model to use for the request. If not provided, defaults to the environment's default model.
    - few_shot (bool, optional): Whether to enable few-shot learning. Defaults to False or the environment setting.
    - few_shot_count (int, optional): The number of few-shot examples to use. Only relevant if few_shot is True.
    - cache_config (CacheConfiguration, optional): Configuration for caching API responses.

    Returns:
    - A decorated function that, when called, sets up the necessary OpperAI function and invokes the API, handling both synchronous and asynchronous execution as needed.

    Examples:
    >>> from opperai import fn
    >>> from pydantic import BaseModel

    >>> class Data(BaseModel):
    >>>     keywords: str
    >>>     sentiment: str

    >>> @fn
    >>> def extract(text = str) -> Data:
    >>>     ''' Extract keywords and sentiment from text. '''

    >>> result = extract(text="Opper is a lovely API that wraps the world of AI")
    >>> print(results)
    """

    def decorator(func):
        func_path = path or func.__name__
        setup_done = False
        sync_client = None
        c = None

        def _prepare_function():
            use_few_shot = few_shot or os.environ.get("OPPER_USE_FEW_SHOT", False)
            function = Function(
                path=func_path,
                description=func.__doc__,
                instructions=f"Operation: {func.__name__}\n\nOperation description: {func.__doc__}",
                out_schema=get_output_schema(func),
                few_shot=use_few_shot,
            )
            if use_few_shot:
                function.use_semantic_search = True
                function.few_shot_count = few_shot_count or 2
            if cache_config:
                function.cache_configuration = cache_config

            function.model = (
                model if model else os.environ.get("OPPER_DEFAULT_MODEL", None)
            )

            return function

        def setup():
            nonlocal setup_done, sync_client, c
            if setup_done:
                return

            if isinstance(client, AsyncClient):
                sync_client = Client(api_key=client.api_key, api_url=client.api_url)
            elif isinstance(client, Client):
                sync_client = client
            else:
                sync_client = Client()

            sync_client.functions.create(_prepare_function())

            if asyncio.iscoroutinefunction(func):
                c = AsyncClient() or client
            else:
                c = sync_client or client

            setup_done = True

        async def async_setup():
            nonlocal setup_done, sync_client, c
            if setup_done:
                return

            if isinstance(client, AsyncClient):
                sync_client = client
            elif isinstance(client, Client):
                sync_client = AsyncClient(
                    api_key=client.api_key, api_url=client.api_url
                )
            else:
                sync_client = AsyncClient()

            await sync_client.functions.create(_prepare_function())

            c = client or sync_client

            setup_done = True

        def _unmarshal_response(response, return_type):
            if return_type is not None:
                if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                    response = return_type(**response)
                elif (
                    (get_origin(return_type) == list or get_origin(return_type) == List)
                    and inspect.isclass(get_args(return_type)[0])
                    and issubclass(get_args(return_type)[0], BaseModel)
                ):
                    response = [get_args(return_type)[0](**item) for item in response]
            return response

        def _prepare_payload(input):
            return ChatPayload(
                parent_span_uuid=get_current_span_id(),
                messages=[
                    Message(role="user", content=json.dumps(input, cls=json_encoder))
                ],
            )

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            await async_setup()

            input = convert_function_call_to_json(func, *args, **kwargs)
            _response = await c.functions.chat(func_path, _prepare_payload(input))

            _thread_local.span_id = _response.span_id

            return_type = get_type_hints(func).get("return")
            answer = _unmarshal_response(_response.json_payload, return_type)

            response = AsyncFunctionResponse(client=c, **_response.model_dump())

            return answer, response

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            setup()

            input = convert_function_call_to_json(func, *args, **kwargs)
            _response = c.functions.chat(func_path, _prepare_payload(input))

            _thread_local.span_id = _response.span_id

            return_type = get_type_hints(func).get("return")
            answer = _unmarshal_response(_response.json_payload, return_type)

            response = FunctionResponse(client=c, **_response.model_dump())

            return answer, response

        return (
            AsyncFunctionWrapper(async_wrapper)
            if asyncio.iscoroutinefunction(func)
            else FunctionWrapper(sync_wrapper)
        )

    if _func is None:
        return decorator
    else:
        return decorator(_func)



from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime

from opperai._client import Client
from opperai.types.spans import Span as SpanModel
from opperai.types.spans import SpanMetric


@dataclass
class Span:
    _client: Client
    uuid: str

    def update(self, **kwargs) -> "Span":
        self._client.spans.update(span_uuid=self.uuid, **kwargs)
        return self

    def delete(self) -> bool:
        return self._client.spans.delete(span_uuid=self.uuid)

    def save_example(self, **kwargs) -> str:
        return self._client.spans.save_example(uuid=self.uuid, **kwargs)

    def save_metric(self, dimension: str, value: float, comment: str = None) -> bool:
        return self._client.spans.save_metric(
            uuid=self.uuid,
            metric=SpanMetric(
                dimension=dimension,
                value=value,
                comment=comment,
            ),
        )


class Spans:
    _client: Client = None

    def __init__(self, client: Client = None):
        if client is None:
            client = Client()

        self._client = client

    @contextmanager
    def start(
        self,
        project: str,
        name: str,
        input: str = None,
        meta: dict = None,
        parent_span_id: str = None,
    ) -> Span:
        span_uuid = self._client.spans.create(
            SpanModel(
                project=project,
                name=name,
                input=input,
                start_time=datetime.now(),
                parent_uuid=parent_span_id,
                meta=meta,
            )
        )
        span = Span(self._client, span_uuid)
        yield span
        span.update(end_time=datetime.now())



# ruff: noqa: F401
from opperai.core.spans._decorator import start_span, trace
from opperai.functions.async_functions import AsyncFunctions
from opperai.functions.decorator._decorator import fn, get_last_span_id
from opperai.functions.functions import Functions
from opperai.indexes.async_indexes import AsyncIndexes
from opperai.indexes.indexes import Indexes
from opperai.spans.async_spans import AsyncSpans
from opperai.spans.spans import Spans

from ._client import AsyncClient, Client


class Opper:
    client: Client = None

    functions: Functions = None
    indexes: Indexes = None
    spans: Spans = None

    def __init__(self, client: Client = None):
        if client is not None:
            if not isinstance(client, Client):
                raise ValueError("Client must be an instance of Client")
        if client is None:
            client = Client()

        self.client = client
        self.functions = Functions(client)
        self.indexes = Indexes(client)
        self.spans = Spans(client)


class AsyncOpper(Opper):
    client: AsyncClient = None

    functions: AsyncFunctions = None
    indexes: AsyncIndexes = None
    spans: AsyncSpans = None

    def __init__(self, client: AsyncClient = None):
        if client is not None:
            if not isinstance(client, AsyncClient):
                raise ValueError("Client must be an instance of AsyncClient")
        if client is None:
            client = AsyncClient()

        self.client = client
        self.functions = AsyncFunctions(client)
        self.indexes = AsyncIndexes(client)
        self.spans = AsyncSpans(client)



from dataclasses import dataclass
from typing import List, Optional

from opperai._client import AsyncClient
from opperai.types.indexes import Document, Filter, RetrievalResponse
from opperai.types.indexes import Index as IndexModel


@dataclass
class AsyncIndex:
    _client: AsyncClient
    _index: IndexModel

    async def upload_file(self, file_path: str, **kwargs):
        """Upload a file to the index."""
        return await self._client.indexes.upload_file(
            id=self._index.id, file_path=file_path, **kwargs
        )

    async def add(self, doc: Document) -> Document:
        """Index a document."""
        return await self._client.indexes.index(id=self._index.id, doc=doc)

    async def query(
        self, query: str, k: int = 10, filters: List[Filter] = None
    ) -> List[RetrievalResponse]:
        """Retrieve documents from the index."""

        return await self._client.indexes.retrieve(
            id=self._index.id, query=query, k=k, filters=filters
        )

    async def delete(self) -> bool:
        """Delete the index."""
        return await self._client.indexes.delete(id=self._index.id)


class AsyncIndexes:
    _client: AsyncClient = None

    def __init__(self, client: AsyncClient = None):
        if client is None:
            client = AsyncClient()

        self._client = client

    async def create(self, name: str) -> AsyncIndex:
        """Create an index with the given name.

        If an index with the given name already exists, return it.
        """
        try:
            index = await self.get(name=name)
            if index:
                return index
        except Exception:
            pass

        index = await self._client.indexes.create(name=name)
        return AsyncIndex(self._client, index)

    async def get(self, id: int = None, name: str = None) -> Optional[AsyncIndex]:
        """Get an index by id or name."""
        if id is not None:
            index = await self._client.indexes.get(id=id)
        elif name is not None:
            index = await self._client.indexes.get(name=name)
        else:
            raise ValueError("Either id or name must be provided")

        if not index:
            return None

        return AsyncIndex(self._client, index)

    async def delete(self, id: int) -> bool:
        """Delete an index by id."""
        return await self._client.indexes.delete(id=id)

    async def list(self) -> List[AsyncIndex]:
        """List all indexes for the organization owning the API key."""
        indexes = await self._client.indexes.list()
        return [AsyncIndex(self._client, index) for index in indexes]
